/* Copyright (c) 2012-2014, The Linux Foundation. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 and
 * only version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/err.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/platform_device.h>
#include <linux/io.h>
#include <linux/i2c/i2c-qup.h>
#include <linux/gpio.h>
#include <linux/irq.h>
#include <linux/irqdomain.h>
#include <linux/of.h>
#include <linux/of_address.h>
#include <linux/of_platform.h>
#include <linux/of_fdt.h>
#include <linux/of_irq.h>
#include <linux/memory.h>
#include <linux/regulator/cpr-regulator.h>
#include <linux/regulator/fan53555.h>
#include <linux/regulator/onsemi-ncp6335d.h>
#include <linux/regulator/qpnp-regulator.h>
#include <linux/regulator/rpm-smd-regulator.h>
#include <linux/msm_tsens.h>
#include <linux/clk/msm-clk-provider.h>
#include <asm/mach/map.h>
#include <asm/mach/arch.h>
#include <asm/mach/time.h>
#include <mach/board.h>
#include <linux/msm-bus.h>
#include <mach/gpiomux.h>
#include <mach/msm_iomap.h>
#ifdef CONFIG_SEC_DEBUG
#include <mach/sec_debug.h>
#endif
#include <mach/msm_memtypes.h>
#include <soc/qcom/restart.h>
#include <soc/qcom/socinfo.h>
#include <mach/board.h>
#include <soc/qcom/smd.h>
#include <soc/qcom/rpm-smd.h>
#include <soc/qcom/smem.h>
#include <linux/msm_thermal.h>
#include "../board-dt.h"
#include "../clock.h"
#include "../platsmp.h"
#include <soc/qcom/spm.h>
#include <soc/qcom/pm.h>
#include <linux/regulator/spm-regulator.h>
#include <linux/platform_data/android_battery.h>
#include <asm/setup.h>
#ifdef CONFIG_SEC_DEBUG
#include <mach/sec_debug.h>
#endif
#if defined (CONFIG_PSTORE_CONSOLE) && (CONFIG_PSTORE_RAM)
#include <linux/pstore_ram.h>
#include <linux/memblock.h>
#endif
struct class *sec_class;
EXPORT_SYMBOL(sec_class);

#if defined (CONFIG_PSTORE_CONSOLE) && (CONFIG_PSTORE_RAM)

static struct ramoops_platform_data ramoops_data = {
	.console_size		= 512 * SZ_1K,
	.mem_size		= SZ_1M,
};

static struct platform_device ramoops_device = {
	.name = "ramoops",
	.dev = {
		.platform_data = &ramoops_data,
	},
};

static void __init sprat_reserve_persist_ram_area(void)
{

	if (memblock.memory.cnt < 2) {
		pr_err("%s: Highmem is not present.\n",
			__func__);
                return;
        }

	ramoops_data.mem_address = memblock.memory.regions[1].base;

	if (memblock_reserve(ramoops_data.mem_address, ramoops_data.mem_size)) {
		pr_err("Failed to reserve persistent memory from %08lx-%08lx\n",
			(long)ramoops_data.mem_address,
			(long)(ramoops_data.mem_address +
				ramoops_data.mem_size - 1));
		return;
	}
	pr_info("ramoops: reserved %d bytes at 0x%08x\n",
		(int)ramoops_data.mem_size, (int)(ramoops_data.mem_address));
}
#else
static void __init sprat_reserve_persist_ram_area(void)
{
	return;
}

#endif

void __init add_persistent_device(void)
{
#if defined (CONFIG_PSTORE_CONSOLE) && (CONFIG_PSTORE_RAM)
	int ret = platform_device_register(&ramoops_device);
	if (ret) {
		pr_err("Unable to register ramoops device (%d)\n", ret);
	}
#endif
}

static void samsung_sys_class_init(void)
{
	pr_info("samsung sys class init.\n");

	sec_class = class_create(THIS_MODULE, "sec");

	if (IS_ERR(sec_class)) {
		pr_err("Failed to create class(sec)!\n");
		return;
	}

	pr_info("samsung sys class end.\n");
};

#if defined(CONFIG_BATTERY_SAMSUNG)
extern void __init samsung_init_battery(void);
#endif

static struct of_dev_auxdata msm_hsic_host_adata[] = {
	OF_DEV_AUXDATA("qcom,hsic-host", 0xF9A00000, "msm_hsic_host", NULL),
	{}
};

#if defined(CONFIG_BATTERY_ANDROID)
extern struct android_bat_platform_data android_battery_pdata;
#endif

static struct of_dev_auxdata msm8226_auxdata_lookup[] __initdata = {
	OF_DEV_AUXDATA("qcom,msm-sdcc", 0xF9824000, \
			"msm_sdcc.1", NULL),
	OF_DEV_AUXDATA("qcom,msm-sdcc", 0xF98A4000, \
			"msm_sdcc.2", NULL),
	OF_DEV_AUXDATA("qcom,msm-sdcc", 0xF9864000, \
			"msm_sdcc.3", NULL),
	OF_DEV_AUXDATA("qcom,sdhci-msm", 0xF9824900, \
			"msm_sdcc.1", NULL),
	OF_DEV_AUXDATA("qcom,sdhci-msm", 0xF98A4900, \
			"msm_sdcc.2", NULL),
	OF_DEV_AUXDATA("qcom,sdhci-msm", 0xF9864900, \
			"msm_sdcc.3", NULL),
	OF_DEV_AUXDATA("qcom,hsic-host", 0xF9A00000, "msm_hsic_host", NULL),
	OF_DEV_AUXDATA("qcom,hsic-smsc-hub", 0, "msm_smsc_hub",
			msm_hsic_host_adata),
#if defined(CONFIG_BATTERY_ANDROID)
	OF_DEV_AUXDATA("android,battery", 0, \
			"android-battery", &android_battery_pdata),
#endif
	{}
};

static void __init msm8226_reserve(void)
{
	of_scan_flat_dt(dt_scan_for_memory_reserve, NULL);
	sprat_reserve_persist_ram_area();
}

/*
 * Used to satisfy dependencies for devices that need to be
 * run early or in a particular order. Most likely your device doesn't fall
 * into this category, and thus the driver should not be added here. The
 * EPROBE_DEFER can satisfy most dependency problems.
 */
void __init msm8226_add_drivers(void)
{
	msm_smd_init();
	msm_rpm_driver_init();
	msm_spm_device_init();
	msm_pm_sleep_status_init();
	rpm_smd_regulator_driver_init();
	qpnp_regulator_init();
	spm_regulator_init();
	msm_gcc_8226_init();
	msm_bus_fabric_init_driver();
	qup_i2c_init_driver();
	ncp6335d_regulator_init();
	fan53555_regulator_init();
	cpr_regulator_init();
	tsens_tm_init_driver();
	msm_thermal_device_init();
	add_persistent_device();
}

void __init msm8226_init(void)
{
	struct of_dev_auxdata *adata = msm8226_auxdata_lookup;

#ifdef CONFIG_SEC_DEBUG
	sec_debug_init();
#endif
	/*
	 * populate devices from DT first so smem probe will get called as part
	 * of msm_smem_init.  socinfo_init needs smem support so call
	 * msm_smem_init before it.  msm8226_init_gpiomux needs socinfo so
	 * call socinfo_init before it.
	 */
	board_dt_populate(adata);

	samsung_sys_class_init();

	msm_smem_init();

	if (socinfo_init() < 0)
		pr_err("%s: socinfo_init() failed\n", __func__);

	msm8226_init_gpiomux();
	msm8226_add_drivers();
}

static const char *msm8226_dt_match[] __initconst = {
	"qcom,msm8226",
	"qcom,msm8926",
	"qcom,apq8026",
	NULL
};

DT_MACHINE_START(MSM8226_DT, "Qualcomm MSM 8226 (Flattened Device Tree)")
	.map_io			= msm_map_msm8226_io,
	.init_machine		= msm8226_init,
	.dt_compat		= msm8226_dt_match,
	.reserve		= msm8226_reserve,
	.smp			= &arm_smp_ops,
MACHINE_END
